# Copyright 2020 Silicon Compiler Authors. All Rights Reserved.

import json
import os
import requests
import shutil
import time
import urllib.parse
import tarfile
import tempfile
import multiprocessing

from siliconcompiler import utils, SiliconCompilerError
from siliconcompiler import NodeStatus as SCNodeStatus
from siliconcompiler._metadata import default_server
from siliconcompiler.flowgraph import nodes_to_execute
from siliconcompiler.remote import JobStatus

# Step name to use while logging
remote_step_name = 'remote'


class Client():
    # Step name to use while logging
    STEP_NAME = "remote"

    def __init__(self, chip, default_server=default_server):
        self.__chip = chip
        self.__logger = self.__chip.logger.getChild('remote-client')

        self.__default_server = default_server

        # Used when reporting node information during run
        self.__maxlinelength = 70

        self.__init_config()
        self.__init_baseurl()

        # Client / server timeout
        self.__timeout = 10
        self.__max_timeouts = 10
        self.__tos_str = '''Please review the SiliconCompiler cloud's terms of service:

https://www.siliconcompiler.com/terms

In particular, please ensure that you have the right to distribute any IP
which is contained in designs that you upload to the service. This public
service, provided by SiliconCompiler, is not intended to process proprietary IP.
'''
        # Runtime
        self.__download_pool = None
        self.__check_interval = None
        self.__node_information = None

    def __get_remote_config_file(self, fail=True):
        if self.__chip.get('option', 'credentials'):
            # Use the provided remote credentials file.
            cfg_file = os.path.abspath(self.__chip.get('option', 'credentials'))

            if fail and not os.path.isfile(cfg_file) and \
               getattr(self, '_error_on_missing_file', True):
                # Check if it's a file since its been requested by the user
                raise SiliconCompilerError(
                    f'Unable to find the credentials file: {cfg_file}',
                    chip=self.__chip)
        else:
            # Use the default config file path.
            cfg_file = utils.default_credentials_file()

        return cfg_file

    def __init_config(self):
        cfg_file = self.__get_remote_config_file()

        remote_cfg = {}
        cfg_dir = os.path.dirname(cfg_file)
        if os.path.isdir(cfg_dir) and os.path.isfile(cfg_file):
            self.__logger.info(f'Using credentials: {cfg_file}')
            with open(cfg_file, 'r') as cfgf:
                remote_cfg = json.loads(cfgf.read())
        else:
            if getattr(self, '_print_server_warning', True):
                self.__logger.warning('Could not find remote server configuration: '
                                      f'defaulting to {self.__default_server}')
            remote_cfg = {
                "address": self.__default_server,
                "directory_whitelist": []
            }
        if 'address' not in remote_cfg:
            raise SiliconCompilerError(
                'Improperly formatted remote server configuration - '
                'please run "sc-remote -configure" and enter your server address and '
                'credentials.', chip=self.__chip)

        self.__config = remote_cfg

    def __init_baseurl(self):
        remote_host = self.__config['address']
        if 'port' in self.__config:
            remote_port = self.__config['port']
        else:
            remote_port = 443
        remote_host += f':{remote_port}'
        if remote_host.startswith('http'):
            remote_protocol = ''
        else:
            remote_protocol = 'https://' if remote_port == 443 else 'http://'
        self.__url = remote_protocol + remote_host

    def __get_url(self, action):
        return urllib.parse.urljoin(self.__url, action)

    def remote_manifest(self):
        return f'{self.__chip.getworkdir()}/sc_remote.pkg.json'

    def print_configuration(self):
        self.__logger.info(f'Server: {self.__url}')
        if 'username' in self.__config:
            self.__logger.info(f'Username: {self.__config["username"]}')
        if 'directory_whitelist' in self.__config and self.__config['directory_whitelist']:
            self.__logger.info('Directory whitelist:')
            for path in sorted(self.__config['directory_whitelist']):
                self.__logger.info(f'  {path}')

    def __get_post_params(self, include_job_name=False, include_job_id=False):
        '''
        Helper function to build the params for the post request
        '''
        # Use authentication if necessary.
        post_params = {}

        if include_job_id and self.__chip.get('record', 'remoteid'):
            post_params['job_hash'] = self.__chip.get('record', 'remoteid')

        if include_job_name and self.__chip.get('option', 'jobname'):
            post_params['job_id'] = self.__chip.get('option', 'jobname')

        # Forward authentication information
        if ('username' in self.__config) and ('password' in self.__config) and \
           (self.__config['username']) and (self.__config['password']):
            post_params['username'] = self.__config['username']
            post_params['key'] = self.__config['password']

        return post_params

    ###################################
    def __post(self, action, post_action, success_action, error_action=None):
        '''
        Helper function to handle the post request
        '''
        redirect_url = self.__get_url(action)

        timeouts = 0
        while redirect_url:
            try:
                resp = post_action(redirect_url)
            except requests.Timeout:
                timeouts += 1
                if timeouts > self.__max_timeouts:
                    raise SiliconCompilerError('Server communications timed out', chip=self.__chip)
                time.sleep(self.__timeout)
                continue
            except Exception as e:
                raise SiliconCompilerError(f'Server communications error: {e}', chip=self.__chip)

            code = resp.status_code
            if 200 <= code and code < 300:
                return success_action(resp)

            try:
                msg_json = resp.json()
                if 'message' in msg_json:
                    msg = msg_json['message']
                else:
                    msg = resp.text
            except requests.JSONDecodeError:
                msg = resp.text

            if 300 <= code and code < 400:
                if 'Location' in resp.headers:
                    redirect_url = resp.headers['Location']
                    continue

            if error_action:
                return error_action(code, msg)
            else:
                raise SiliconCompilerError(f'Server responded with {code}: {msg}', chip=self.__chip)

    def cancel_job(self):
        '''
        Helper method to request that the server cancel an ongoing job.
        '''

        def post_action(url):
            return requests.post(
                url,
                data=json.dumps(self.__get_post_params(
                    include_job_id=True
                )),
                timeout=self.__timeout)

        def success_action(resp):
            return json.loads(resp.text)

        return self.__post('/cancel_job/', post_action, success_action)

    ###################################
    def delete_job(self):
        '''
        Helper method to delete a job from shared remote storage.
        '''

        def post_action(url):
            return requests.post(
                url,
                data=json.dumps(self.__get_post_params(
                    include_job_id=True
                )),
                timeout=self.__timeout)

        def success_action(resp):
            return resp.text

        return self.__post('/delete_job/', post_action, success_action)

    def check_job_status(self):
        # Make the request and print its response.
        def post_action(url):
            return requests.post(
                url,
                data=json.dumps(self.__get_post_params(include_job_id=True, include_job_name=True)),
                timeout=self.__timeout)

        def error_action(code, msg):
            return {
                'busy': True,
                'message': ''
            }

        def success_action(resp):
            json_response = json.loads(resp.text)

            if json_response['status'] != JobStatus.RUNNING:
                if json_response['status'] == JobStatus.REJECTED:
                    self.__logger.error(f'Job was rejected: {json_response["message"]}')
                elif json_response['status'] != JobStatus.UNKNOWN:
                    self.__logger.info(f'Job status: {json_response["status"]}')
            info = {
                'busy': json_response['status'] == JobStatus.RUNNING,
                'message': None
            }

            if isinstance(json_response['message'], str):
                info['message'] = json_response['message']
            else:
                info['message'] = json.dumps(json_response['message'])
            return info

        info = self.__post(
            '/check_progress/',
            post_action,
            success_action,
            error_action=error_action)

        if not info:
            info = {
                'busy': True,
                'message': ''
            }
        return info

    def __log_node_status(self, status, nodes):
        '''
        Helper method to log truncated information about flowgraph nodes
        with a given status, on a single line.
        Used to print info about all statuses besides NodeStatus.RUNNING.
        '''

        num_nodes = len(nodes)
        if num_nodes > 0:
            line_len = 0
            nodes_log = f'  {status.title()} ({num_nodes}): '
            log_nodes = []
            for node, _ in nodes:
                node_len = len(node)

                if node_len + line_len + 2 < self.__maxlinelength:
                    log_nodes.append(node)
                    line_len += node_len + 2
                else:
                    if len(log_nodes) == num_nodes - 1:
                        log_nodes.append(node)
                    else:
                        log_nodes.append('...')
                    break

            nodes_log += ', '.join(log_nodes)
            self.__logger.info(nodes_log)

    def _report_job_status(self, info):
        if not info['busy']:
            # Job is not running
            return [], False

        try:
            # Decode response JSON, if possible.
            job_info = json.loads(info['message'])
        except json.JSONDecodeError as e:
            self.__logger.warning(f"Job is still running: {e}")
            return [], True

        completed = []

        nodes_to_log = {}
        for node, node_info in job_info.items():
            status = node_info['status']
            nodes_to_log.setdefault(status, []).append((node, node_info))

            if SCNodeStatus.is_done(status):
                # collect completed
                completed.append(node)

            if self.__node_information and node in self.__node_information:
                self.__chip.set('record', 'status', status,
                                step=self.__node_information[node]["step"],
                                index=self.__node_information[node]["index"])

        nodes_to_log = {key: nodes_to_log[key] for key in sorted(nodes_to_log.keys())}

        # Log information about the job's progress.
        self.__logger.info("Job is still running. Status:")

        for stat, nodes in nodes_to_log.items():
            if SCNodeStatus.is_done(stat):
                self.__log_node_status(stat, nodes)

        # Running / in-progress flowgraph nodes should all be printed:
        for stat, nodes in nodes_to_log.items():
            if SCNodeStatus.is_running(stat):
                self.__logger.info(f'  {stat.title()} ({len(nodes)}):')
                for node, node_info in nodes:
                    running_log = f"    {node}"
                    if 'elapsed_time' in node_info:
                        running_log += f" ({node_info['elapsed_time']})"
                    self.__logger.info(running_log)

        # Queued and pending flowgraph nodes:
        for stat, nodes in nodes_to_log.items():
            if SCNodeStatus.is_waiting(stat):
                self.__log_node_status(stat, nodes)

        return completed, True

    def __check(self):
        def post_action(url):
            return requests.post(
                url,
                data=json.dumps(self.__get_post_params()),
                timeout=self.__timeout)

        def success_action(resp):
            return resp.json()

        response_info = self.__post('/check_server/', post_action, success_action)
        if not response_info:
            raise ValueError('Server response is not valid.')

        return response_info

    def check(self):
        '''
        Helper method to call check_server on server
        '''

        # Make the request and print its response.
        response_info = self.__check()

        # Print status value.
        server_status = response_info['status']
        self.__logger.info(f'Server status: {server_status}')
        if server_status != 'ready':
            self.__logger.warning('  Status is not "ready", server cannot accept new jobs.')

        # Print server-side version info.
        version_info = response_info['versions']
        version_suffix = ' version'
        max_name_string_len = max([len(s) for s in version_info.keys()]) + len(version_suffix)
        self.__logger.info('Server software versions:')
        for name, version in version_info.items():
            print_name = f'{name}{version_suffix}'
            self.__logger.info(f'  {print_name: <{max_name_string_len}}: {version}')

        # Print user info if applicable.
        if 'user_info' in response_info:
            user_info = response_info['user_info']
            if ('compute_time' not in user_info) or \
               ('bandwidth_kb' not in user_info):
                self.__logger.info('Error fetching user information from the remote server.')
                raise ValueError(f'Server response is not valid or missing fields: {user_info}')

            if 'username' in self.__config:
                # Print the user's account info, and return.
                self.__logger.info(f'User {self.__config["username"]}:')

            time_remaining = user_info["compute_time"] / 60.0
            bandwidth_remaining = user_info["bandwidth_kb"]
            self.__logger.info(f'  Remaining compute time: {(time_remaining):.2f} minutes')
            self.__logger.info(f'  Remaining results bandwidth: {bandwidth_remaining} KiB')

        self.__print_tos(response_info)

        # Return the response info in case the caller wants to inspect it.
        return response_info

    def __print_tos(self, response_info):
        # Print terms-of-service message, if the server provides one.
        if 'terms' in response_info and response_info['terms']:
            self.__logger.info('Terms of Service info for this server:')
            for line in response_info['terms'].splitlines():
                if line:
                    self.__logger.info(line)

    def run(self):
        '''
        Dispatch the Chip to a remote server for processing.
        '''
        should_resume = not self.__chip.get('option', 'clean')
        remote_resume = should_resume and self.__chip.get('record', 'remoteid')

        # Pre-process: Run an starting nodes locally, and upload the
        # in-progress build directory to the remote server.
        # Data is encrypted if user / key were specified.
        # run remote process
        if self.__chip.get('arg', 'step'):
            raise SiliconCompilerError('Cannot pass [arg,step] parameter into remote flow.',
                                       chip=self.__chip)
        if self.__chip.get('arg', 'index'):
            raise SiliconCompilerError('Cannot pass [arg,index] parameter into remote flow.',
                                       chip=self.__chip)

        # Only run the pre-process step if the job doesn't already have a remote ID.
        if not remote_resume:
            self.__run_preprocess()

        # Run the job on the remote server, and wait for it to finish.
        # Set logger to indicate remote run
        self.__chip._init_logger(step=self.STEP_NAME, index=None, in_run=True)

        # Ask the remote server to start processing the requested step.
        self.__request_run()

        # Run the main 'check_progress' loop to monitor job status until it finishes.
        self._run_loop()

        # Restore logger
        self.__chip._init_logger(in_run=True)

    def __request_run(self):
        '''
        Helper method to make a web request to start a job stage.
        '''

        remote_status = self.__check()

        if remote_status['status'] != 'ready':
            raise SiliconCompilerError('Remote server is not available.', chip=self.__chip)

        self.__print_tos(remote_status)

        remote_resume = not self.__chip.get('option', 'clean') and \
            self.__chip.get('record', 'remoteid')
        # Only package and upload the entry steps if starting a new job.
        if not remote_resume:
            upload_file = tempfile.TemporaryFile(prefix='sc', suffix='remote.tar.gz')
            with tarfile.open(fileobj=upload_file, mode='w:gz') as tar:
                tar.add(self.__chip.getworkdir(), arcname='')
            # Flush file to ensure everything is written
            upload_file.flush()

            # We no longer need the collected files
            shutil.rmtree(self.__chip._getcollectdir(jobname=self.__chip.get('option', 'jobname')))

        if 'pre_upload' in remote_status:
            self.__logger.info(remote_status['pre_upload']['message'])
            time.sleep(remote_status['pre_upload']['delay'])

        # Make the actual request, streaming the bulk data as a multipart file.
        # Redirected POST requests are translated to GETs. This is actually
        # part of the HTTP spec, so we need to manually follow the trail.
        post_params = {
            'chip_cfg': self.__chip.schema.cfg,
            'params': self.__get_post_params(include_job_id=True)
        }

        post_files = {'params': json.dumps(post_params)}
        if not remote_resume:
            post_files['import'] = upload_file
            upload_file.seek(0)

        def post_action(url):
            return requests.post(
                url,
                files=post_files,
                timeout=self.__timeout)

        def success_action(resp):
            return resp.json()

        resp = self.__post('/remote_run/', post_action, success_action)
        if not remote_resume:
            upload_file.close()

        if 'message' in resp and resp['message']:
            self.__logger.info(resp['message'])
        self.__chip.set('record', 'remoteid', resp['job_hash'])

        self.__chip.write_manifest(self.remote_manifest())

        self.__logger.info(f"Your job's reference ID is: {resp['job_hash']}")

        self.__check_interval = remote_status['progress_interval']

    def __run_preprocess(self):
        '''
        Helper method to run a local import stage for remote jobs.
        '''

        # Ensure packages with python sources are copied
        for key in self.__chip.allkeys():
            key_type = self.__chip.get(*key, field='type')

            if 'dir' in key_type or 'file' in key_type:
                for _, step, index in self.__chip.schema._getvals(*key, return_defvalue=False):
                    packages = self.__chip.get(*key, field='package', step=step, index=index)
                    force_copy = False
                    for package in packages:
                        if not package:
                            continue
                        if package.startswith('python://'):
                            force_copy = True
                    if force_copy:
                        self.__chip.set(*key, True, field='copy', step=step, index=index)

        # Collect inputs into a collection directory only for remote runs, since
        # we need to send inputs up to the server.
        self.__chip.collect(whitelist=self.__config.setdefault('directory_whitelist', []))

    def _run_loop(self):
        # Wrapper to allow for capturing of Ctrl+C
        try:
            self.__run_loop()
            self._finalize_loop()
        except KeyboardInterrupt:
            manifest_path = self.remote_manifest()
            reconnect_cmd = f'sc-remote -cfg {manifest_path} -reconnect'
            cancel_cmd = f'sc-remote -cfg {manifest_path} -cancel'
            self.__logger.info('Disconnecting from remote job')
            self.__logger.info(f'To reconnect to this job use: {reconnect_cmd}')
            self.__logger.info(f'To cancel this job use: {cancel_cmd}')
            raise SiliconCompilerError('Job canceled by user keyboard interrupt')

    def __import_run_manifests(self):
        changed = False
        for _, node_info in self.__node_information.items():
            if node_info["imported"]:
                continue

            manifest = os.path.join(
                self.__chip.getworkdir(step=node_info["step"], index=node_info["index"]),
                'outputs',
                f'{self.__chip.design}.pkg.json')
            if os.path.exists(manifest):
                try:
                    self.__chip.schema.read_journal(manifest)
                    node_info["imported"] = True
                    changed = True
                except:  # noqa E722
                    # Import may fail if file is still getting written
                    pass
        return changed

    def __ensure_run_loop_information(self):
        self.__chip._init_logger(step=self.STEP_NAME, index='0', in_run=True)
        if not self.__download_pool:
            self.__download_pool = multiprocessing.Pool()

        if self.__check_interval is None:
            check_info = self.__check()
            self.__check_interval = check_info['progress_interval']

        self.__node_information = {}
        for step, index in nodes_to_execute(self.__chip):
            done = SCNodeStatus.is_done(self.__chip.get('record', 'status', step=step, index=index))
            node_info = {
                "step": step,
                "index": index,
                "imported": done,
                "fetched": done
            }
            self.__node_information[f'{step}{index}'] = node_info

    def __run_loop(self):
        self.__ensure_run_loop_information()

        # Check the job's progress periodically until it finishes.
        running = True

        while running:
            time.sleep(self.__check_interval)
            self.__import_run_manifests()

            # Check progress
            job_info = self.check_job_status()
            completed, running = self._report_job_status(job_info)

            if self.__chip._dash:
                # Update dashboard if active
                self.__chip._dash.update_manifest()

            nodes_to_fetch = []
            for node in completed:
                if not self.__node_information[node]["fetched"]:
                    nodes_to_fetch.append(node)

            if nodes_to_fetch:
                self.__logger.info('  Fetching completed results:')
                for node in nodes_to_fetch:
                    self.__schedule_fetch_result(node)

        # Done: try to fetch any node results which still haven't been retrieved.
        self.__logger.info('Remote job completed! Retrieving final results...')
        for node, node_info in self.__node_information.items():
            if not node_info["fetched"]:
                self.__schedule_fetch_result(node)
        self.__schedule_fetch_result(node)

        self._finalize_loop()

        # Un-set the 'remote' option to avoid from/to-based summary/show errors
        self.__chip.unset('option', 'remote')

        if self.__chip._dash:
            self.__chip._dash.update_manifest()

    def _finalize_loop(self):
        if self.__download_pool:
            self.__download_pool.close()
            self.__download_pool.join()
            self.__download_pool = None

        self.__import_run_manifests()

    def __schedule_fetch_result(self, node):
        self.__node_information[node]["fetched"] = True
        self.__download_pool.apply_async(Client._fetch_result, (self, node))
        if node is None:
            node = 'final result'
        self.__logger.info(f'    {node}')

    def _fetch_result(self, node):
        '''
        Helper method to fetch and open job results from a remote compute cluster.
        Optional 'node' argument fetches results for only the specified
        flowgraph node (e.g. "floorplan0")
        '''

        # Collect local values.
        job_hash = self.__chip.get('record', 'remoteid')
        local_dir = self.__chip.get('option', 'builddir')

        # Set default results archive path if necessary, and fetch it.
        with tempfile.TemporaryDirectory(prefix=f'sc_{job_hash}_', suffix=f'_{node}') as tmpdir:
            results_path = os.path.join(tmpdir, 'result.tar.gz')

            with open(results_path, 'wb') as rd:
                # Fetch results archive.
                def post_action(url):
                    post_params = self.__get_post_params()
                    if node:
                        post_params['node'] = node
                    return requests.post(
                        url,
                        data=json.dumps(post_params),
                        stream=True,
                        timeout=self.__timeout)

                def success_action(resp):
                    shutil.copyfileobj(resp.raw, rd)
                    return 0

                def error_action(code, msg):
                    # Results are fetched in parallel, and a failure in one node
                    # does not necessarily mean that the whole job failed.
                    if node:
                        self.__logger.warning(f'Could not fetch results for node: {node}')
                    else:
                        self.__logger.warning('Could not fetch results for final results.')
                    return 404

                results_code = self.__post(
                    f'/get_results/{job_hash}.tar.gz',
                    post_action,
                    success_action,
                    error_action=error_action
                )

            # Note: the server should eventually delete the results as they age out (~8h),
            # but this will give us a brief period to look at failed results.
            if results_code:
                return

            # Unzip the results.
            # Unauthenticated jobs get a gzip archive, authenticated jobs get nested archives.
            # So we need to extract and delete those.
            # Archive contents: server-side build directory. Format:
            # [job_hash]/[design]/[job_name]/[step]/[index]/...
            try:
                with tarfile.open(results_path, 'r:gz') as tar:
                    tar.extractall(path=tmpdir)
            except tarfile.TarError as e:
                self.__logger.error(f'Failed to extract data from {results_path}: {e}')
                return

            work_dir = os.path.join(tmpdir, job_hash)
            if os.path.exists(work_dir):
                shutil.copytree(work_dir, local_dir, dirs_exist_ok=True)
            else:
                self.__logger.error(f'Empty file returned from remote for: {node}')
                return

    def configure_server(self, server=None, username=None, password=None):

        def confirm_dialog(message):
            confirmed = False
            while not confirmed:
                oin = input(f'{message} y/N: ')
                if (not oin) or (oin == 'n') or (oin == 'N'):
                    return False
                elif (oin == 'y') or (oin == 'Y'):
                    return True
            return False

        default_server_name = urllib.parse.urlparse(self.__default_server).hostname

        # Find the config file/directory path.
        cfg_file = self.__get_remote_config_file()
        cfg_dir = os.path.dirname(cfg_file)

        # Create directory if it doesn't exist.
        if cfg_dir and not os.path.isdir(cfg_dir):
            os.makedirs(cfg_dir, exist_ok=True)

        # If an existing config file exists, prompt the user to overwrite it.
        if os.path.isfile(cfg_file):
            if not confirm_dialog('Overwrite existing remote configuration?'):
                return

        self.__config = {}

        # If a command-line argument is passed in, use that as a public server address.
        if server:
            srv_addr = server
            self.__logger.info(f'Creating remote configuration file for server: {srv_addr}')
        else:
            # If no arguments were passed in, interactively request credentials from the user.
            srv_addr = input('Remote server address (leave blank to use default server):\n')
            srv_addr = srv_addr.replace(" ", "")

        if not srv_addr:
            srv_addr = self.__default_server
            self.__logger.info(f'Using {srv_addr} as server')

        server = urllib.parse.urlparse(srv_addr)
        has_scheme = True
        if not server.hostname:
            # fake add a scheme to the url
            has_scheme = False
            server = urllib.parse.urlparse('https://' + srv_addr)
        if not server.hostname:
            raise ValueError(f'Invalid address provided: {srv_addr}')

        if has_scheme:
            self.__config['address'] = f'{server.scheme}://{server.hostname}'
        else:
            self.__config['address'] = server.hostname

        public_server = default_server_name in srv_addr
        if public_server and not confirm_dialog(self.__tos_str):
            return

        if server.port is not None:
            self.__config['port'] = server.port

        if not public_server:
            if username is None:
                username = server.username
                if username is None:
                    username = input('Remote username (leave blank for no username):\n')
                    username = username.replace(" ", "")
            if password is None:
                password = server.password
                if password is None:
                    password = input('Remote password (leave blank for no password):\n')
                    password = password.replace(" ", "")

            if username:
                self.__config['username'] = username
            if password:
                self.__config['password'] = password

        self.__config['directory_whitelist'] = []

        # Save the values to the target config file in JSON format.
        with open(cfg_file, 'w') as f:
            f.write(json.dumps(self.__config, indent=4))

        # Let the user know that we finished successfully.
        self.__logger.info(f'Remote configuration saved to: {cfg_file}')

    def configure_whitelist(self, add=None, remove=None):
        cfg_file = self.__get_remote_config_file()

        self.__logger.info(f'Updating credentials: {cfg_file}')

        if 'directory_whitelist' not in self.__config:
            self.__config['directory_whitelist'] = []

        if add:
            for path in add:
                path = os.path.abspath(path)
                self.__logger.info(f'Adding {path}')
                self.__config['directory_whitelist'].append(path)

        if remove:
            for path in remove:
                path = os.path.abspath(path)
                if path in self.__config['directory_whitelist']:
                    self.__logger.info(f'Removing {path}')
                    self.__config['directory_whitelist'].remove(path)

        # Cleanup
        self.__config['directory_whitelist'] = list(set(self.__config['directory_whitelist']))

        # Save the values to the target config file in JSON format.
        with open(cfg_file, 'w') as f:
            f.write(json.dumps(self.__config, indent=4))

    #######################################
    def __getstate__(self):
        # Called when generating a serial stream of the object
        attributes = self.__dict__.copy()

        attributes['_Client__download_pool'] = None

        return attributes


class ConfigureClient(Client):
    def __init__(self, chip):
        self._print_server_warning = False
        self._error_on_missing_file = False

        super().__init__(chip)



import os
import json
from siliconcompiler import sc_open
from siliconcompiler import utils
from siliconcompiler.tools._common import input_provides, add_common_file, \
    get_tool_task, record_metric
from siliconcompiler.tools._common.asic import get_mainlib, set_tool_task_var, get_libraries, \
    CellArea, set_tool_task_lib_var
from siliconcompiler.tools.openroad import setup as tool_setup


def setup(chip, exit=True):
    tool_setup(chip, exit=exit)

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    chip.set('tool', tool, 'task', task, 'threads', utils.get_cores(chip),
             step=step, index=index)

    pdkname = chip.get('option', 'pdk')
    targetlibs = get_libraries(chip, 'logic')
    mainlib = get_mainlib(chip)
    macrolibs = get_libraries(chip, 'macro')
    stackup = chip.get('option', 'stackup')
    delaymodel = chip.get('asic', 'delaymodel', step=step, index=index)
    libtype = chip.get('library', mainlib, 'asic', 'libarch', step=step, index=index)

    if delaymodel != 'nldm':
        chip.error(f'{delaymodel} delay model is not supported by {tool}, only nldm')

    if not stackup or not targetlibs:
        chip.error('Stackup and logiclib parameters required for OpenROAD.')

    # Setup general required
    chip.add('tool', tool, 'task', task, 'require', 'asic,logiclib',
             step=step, index=index)
    if chip.get('asic', 'macrolib', step=step, index=index):
        chip.add('tool', tool, 'task', task, 'require', 'asic,macrolib',
                 step=step, index=index)
    if chip.get('option', 'library', step=step, index=index):
        chip.add('tool', tool, 'task', task, 'require', 'option,library',
                 step=step, index=index)
    chip.add('tool', tool, 'task', task, 'require', 'asic,delaymodel',
             step=step, index=index)
    chip.add('tool', tool, 'task', task, 'require', 'option,stackup',
             step=step, index=index)
    chip.add('tool', tool, 'task', task, 'require', f'library,{mainlib},asic,libarch',
             step=step, index=index)
    chip.add('tool', tool, 'task', task, 'require', f'library,{mainlib},asic,site,{libtype}',
             step=step, index=index)
    chip.add('tool', tool, 'task', task, 'require',
             f'pdk,{pdkname},aprtech,openroad,{stackup},{libtype},lef',
             step=step, index=index)

    # Add library requirements
    for lib in targetlibs:
        for timing_key in get_library_timing_keypaths(chip, lib).values():
            chip.add('tool', tool, 'task', task, 'require', ",".join(timing_key),
                     step=step, index=index)
        chip.add('tool', tool, 'task', task, 'require',
                 ",".join(['library', lib, 'output', stackup, 'lef']),
                 step=step, index=index)
    for lib in macrolibs:
        for timing_key in get_library_timing_keypaths(chip, lib).values():
            if chip.valid(*timing_key):
                chip.add('tool', tool, 'task', task, 'require', ",".join(timing_key),
                         step=step, index=index)
        chip.add('tool', tool, 'task', task, 'require',
                 ",".join(['library', lib, 'output', stackup, 'lef']),
                 step=step, index=index)


def extract_metrics(chip):
    '''
    Extract metrics
    '''

    # Check log file for errors and statistics
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')

    metric_reports = {
        "setuptns": [
            "timing/total_negative_slack.rpt",
            "timing/setup.rpt"
        ],
        "setupslack": [
            "timing/worst_slack.setup.rpt",
            "timing/setup.rpt",
            "timing/setup.topN.rpt"
        ],
        "setupskew": [
            "timing/skew.setup.rpt",
            "timing/worst_slack.setup.rpt",
            "timing/setup.rpt",
            "timing/setup.topN.rpt"
        ],
        "setuppaths": [
            "timing/setup.rpt",
            "timing/setup.topN.rpt"
        ],
        "holdslack": [
            "timing/worst_slack.hold.rpt",
            "timing/hold.rpt",
            "timing/hold.topN.rpt"
        ],
        "holdskew": [
            "timing/skew.hold.rpt",
            "timing/worst_slack.hold.rpt",
            "timing/hold.rpt",
            "timing/hold.topN.rpt"
        ],
        "holdpaths": [
            "timing/hold.rpt",
            "timing/hold.topN.rpt"
        ],
        "unconstrained": [
            "timing/unconstrained.rpt",
            "timing/unconstrained.topN.rpt"
        ],
        "peakpower": [
            *[f"power/{corner}.rpt" for corner in chip.getkeys('constraint', 'timing')],
            *[f"images/heatmap/power_density/{corner}.png"
              for corner in chip.getkeys('constraint', 'timing')]
        ],
        "drvs": [
            "timing/drv_violators.rpt",
            "floating_nets.rpt",
            "overdriven_nets.rpt",
            "overdriven_nets_with_parallel.rpt",
            f"{chip.design}_antenna.rpt",
            f"{chip.design}_antenna_post_repair.rpt"
        ],
        "drcs": [
            f"{chip.design}_drc.rpt",
            f"markers/{chip.design}.drc.rpt",
            f"markers/{chip.design}.drc.json",
            f"images/markers/{chip.design}.DRC.png"
        ],
        "utilization": [
            "images/heatmap/placement_density.png"
        ],
        "wirelength": [
            f"images/{chip.design}.routing.png"
        ]
    }
    metric_reports["leakagepower"] = metric_reports["peakpower"]

    metrics_file = "reports/metrics.json"

    if not os.path.exists(metrics_file):
        chip.logger.warning("OpenROAD metrics file is missing")
        return

    def get_metric_sources(metric):
        metric_sources = [metrics_file]
        if metric in metric_reports:
            for metric_file in metric_reports[metric]:
                metric_path = f'reports/{metric_file}'
                if os.path.exists(metric_path):
                    metric_sources.append(metric_path)
        return metric_sources

    # parsing log file
    with sc_open(metrics_file) as f:
        try:
            metrics = json.load(f)
        except json.decoder.JSONDecodeError as e:
            chip.logger.error(f'Unable to parse metrics from OpenROAD: {e}')
            metrics = {}

        _generate_cell_area_report(chip.top(), metrics)

        or_units = {}
        for unit, or_unit in [
                ('time', 'run__flow__platform__time_units'),
                ('capacitance', 'run__flow__platform__capacitance_units'),
                ('resistance', 'run__flow__platform__resistance_units'),
                ('volt', 'run__flow__platform__voltage_units'),
                ('amp', 'run__flow__platform__current_units'),
                ('power', 'run__flow__platform__power_units'),
                ('distance', 'run__flow__platform__distance_units')]:
            if or_unit in metrics:
                # Remove first digit
                metric_unit = metrics[or_unit][1:]
                or_units[unit] = metric_unit

        or_units['distance'] = 'um'  # always microns
        or_units['power'] = 'W'  # always watts
        or_units['area'] = f"{or_units['distance']}^2"
        or_units['frequency'] = 'Hz'  # always hertz

        has_timing = True
        if 'sc__metric__timing__clocks' in metrics:
            has_timing = metrics['sc__metric__timing__clocks'] > 0

        for metric, or_metric, or_use, or_unit in [
            ('vias', 'sc__step__global_route__vias', True, None),
            ('vias', 'sc__step__route__vias', True, None),
            ('wirelength', 'sc__step__global_route__wirelength', True, 'distance'),
            ('wirelength', 'sc__step__route__wirelength', True, 'distance'),
            ('cellarea', 'sc__metric__design__instance__area', True, 'area'),
            ('stdcellarea', 'sc__metric__design__instance__area__stdcell', True, 'area'),
            ('macroarea', 'sc__metric__design__instance__area__macros', True, 'area'),
            ('padcellarea', 'sc__metric__design__instance__area__padcells', True, 'area'),
            ('totalarea', 'sc__metric__design__die__area', True, 'area'),
            ('utilization', 'sc__metric__design__instance__utilization', True, 100.0),
            ('setuptns', 'sc__metric__timing__setup__tns', has_timing, 'time'),
            ('holdtns', 'sc__metric__timing__hold__tns', has_timing, 'time'),
            ('setupslack', 'sc__metric__timing__setup__ws', has_timing, 'time'),
            ('holdslack', 'sc__metric__timing__hold__ws', has_timing, 'time'),
            ('setupskew', 'sc__metric__clock__skew__setup', has_timing, 'time'),
            ('holdskew', 'sc__metric__clock__skew__hold', has_timing, 'time'),
            ('fmax', 'sc__metric__timing__fmax', has_timing, 'frequency'),
            ('setuppaths', 'sc__metric__timing__drv__setup_violation_count', True, None),
            ('holdpaths', 'sc__metric__timing__drv__hold_violation_count', True, None),
            ('unconstrained', 'sc__metric__timing__unconstrained', True, None),
            ('peakpower', 'sc__metric__power__total', True, 'power'),
            ('leakagepower', 'sc__metric__power__leakage__total', True, 'power'),
            ('pins', 'sc__metric__design__io', True, None),
            ('cells', 'sc__metric__design__instance__count', True, None),
            ('macros', 'sc__metric__design__instance__count__macros', True, None),
            ('nets', 'sc__metric__design__nets', True, None),
            ('registers', 'sc__metric__design__registers', True, None),
            ('inverters', 'sc__metric__design__inverters', True, None),
            ('buffers', 'sc__metric__design__buffers', True, None),
            ('logicdepth', 'sc__metric__design__logic__depth', True, None)
        ]:
            if or_metric in metrics:
                value = metrics[or_metric]

                # Check for INF timing
                if or_unit == 'time' and abs(value) > 1e24:
                    or_use = False

                if or_unit:
                    if or_unit in or_units:
                        or_unit = or_units[or_unit]
                    else:
                        value *= or_unit
                        or_unit = None

                if or_use:
                    record_metric(chip, step, index, metric, value,
                                  get_metric_sources(metric),
                                  source_unit=or_unit)

        ir_drop = None
        for or_metric, value in metrics.items():
            if or_metric.startswith("sc__step__design_powergrid__drop__worst__net") or \
               or_metric.startswith("sc__image__design_powergrid__drop__worst__net"):
                if not ir_drop:
                    ir_drop = value
                else:
                    ir_drop = max(value, ir_drop)

        if ir_drop is not None:
            record_metric(chip, step, index, 'irdrop', ir_drop,
                          get_metric_sources('irdrop'),
                          source_unit='V')

        # setup wns and hold wns can be computed from setup slack and hold slack
        if 'sc__metric__timing__setup__ws' in metrics and \
           has_timing and \
           chip.get('metric', 'setupslack', step=step, index=index) is not None:
            wns = min(0.0, chip.get('metric', 'setupslack', step=step, index=index))
            wns_units = chip.get('metric', 'setupslack', field='unit')
            record_metric(chip, step, index, 'setupwns', wns,
                          get_metric_sources('setupslack'),
                          source_unit=wns_units)

        if 'sc__metric__timing__hold__ws' in metrics and \
           has_timing and \
           chip.get('metric', 'holdslack', step=step, index=index) is not None:
            wns = min(0.0, chip.get('metric', 'holdslack', step=step, index=index))
            wns_units = chip.get('metric', 'holdslack', field='unit')
            record_metric(chip, step, index, 'holdwns', wns,
                          get_metric_sources('holdslack'),
                          source_unit=wns_units)

        drvs = None
        for metric in [
                'sc__metric__timing__drv__max_slew',
                'sc__metric__timing__drv__max_cap',
                'sc__metric__timing__drv__max_fanout',
                'sc__metric__timing__drv__max_fanout',
                'sc__metric__timing__drv__floating__nets',
                'sc__metric__timing__drv__floating__pins',
                'sc__metric__timing__drv__overdriven__nets',
                'sc__metric__antenna__violating__nets',
                'sc__metric__antenna__violating__pins']:
            if metric in metrics:
                if drvs is None:
                    drvs = int(metrics[metric])
                else:
                    drvs += int(metrics[metric])

        if drvs is not None:
            record_metric(chip, step, index, 'drvs', drvs, get_metric_sources('drvs'))

        if 'sc__step__route__drc_errors' in metrics:
            drcs = int(metrics['sc__step__route__drc_errors'])
            record_metric(chip, step, index, 'drcs', drcs, get_metric_sources('drcs'))


def _generate_cell_area_report(design, ord_metrics):
    cellarea_report = CellArea()

    prefix = "sc__cellarea__design__instance"

    filtered_data = {}
    for key, value in ord_metrics.items():
        if key.startswith(prefix):
            filtered_data[key[len(prefix)+2:]] = value

    modules = set()
    modules.add("")
    for key in filtered_data.keys():
        if "__in_module:" in key:
            module = key[key.find("__in_module:"):]
            modules.add(module)

    def process_cell(group):
        data = {}
        for key, value in filtered_data.items():
            if (group != "" and key.endswith(group)):
                key = key[:key.find("__in_module:")]
                data[key] = value
            elif (group == "" and "__in_module" not in key):
                data[key] = value

        cell_type = None
        cell_name = None

        if not group:
            cell_type = design
            cell_name = design
        else:
            cell_type = group[len("__in_module:"):]

        cellarea = None
        cellcount = None

        macroarea = None
        macrocount = None

        stdcell_types = (
            'tie_cell',
            'standard_cell',
            'buffer',
            'clock_buffer',
            'timing_repair_buffer',
            'inverter',
            'clock_inverter',
            'timing_Repair_inverter',
            'clock_gate_cell',
            'level_shifter_cell',
            'sequential_cell',
            'multi_input_combinational_cell',
            'other'
            )

        stdcell_info_area = []
        stdcell_info_count = []
        stdcellarea = None
        stdcellcount = None

        for key, value in data.items():
            if key == 'name':
                cell_name = value
            elif key == 'count':
                cellcount = value
            elif key == 'area':
                cellarea = value
            elif key.startswith('count__class'):
                _, cell_class = key.split(':')
                if cell_class == 'macro':
                    macrocount = value
                elif cell_class in stdcell_types:
                    stdcell_info_count.append(value)
            elif key.startswith('area__class'):
                _, cell_class = key.split(':')
                if cell_class == 'macro':
                    macroarea = value
                elif cell_class in stdcell_types:
                    stdcell_info_area.append(value)

        if stdcell_info_count:
            stdcellcount = sum(stdcell_info_count)
        if stdcell_info_area:
            stdcellarea = sum(stdcell_info_area)

        cellarea_report.addCell(
            name=cell_name,
            module=cell_type,
            cellarea=cellarea,
            cellcount=cellcount,
            macroarea=macroarea,
            macrocount=macrocount,
            stdcellarea=stdcellarea,
            stdcellcount=stdcellcount)

        if filtered_data:
            return True
        return False

    for module in modules:
        process_cell(module)

    if cellarea_report.size() > 0:
        cellarea_report.writeReport("reports/hierarchical_cell_area.json")


def define_tapcell_params(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    # Files
    chip.set('tool', tool, 'task', task, 'file', 'ifp_tapcell',
             'tap cell insertion script',
             field='help')


def define_tapcell_files(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    pdkname = chip.get('option', 'pdk')
    mainlib = get_mainlib(chip)
    stackup = chip.get('option', 'stackup')
    libtype = chip.get('library', mainlib, 'asic', 'libarch',
                       step=step, index=index)

    # set tapcell file
    tapfile = None
    if chip.valid('library', mainlib, 'option', 'file', 'openroad_tapcells'):
        tapfile = chip.find_files('library', mainlib, 'option', 'file', 'openroad_tapcells')
    elif chip.valid('pdk', pdkname, 'aprtech', tool, stackup, libtype, 'tapcells'):
        tapfile = chip.find_files('pdk', pdkname, 'aprtech', tool, stackup, libtype, 'tapcells')
    if tapfile:
        chip.set('tool', tool, 'task', task, 'file', 'ifp_tapcell', tapfile,
                 step=step, index=index, clobber=False)


def define_tiecell_params(chip):
    set_tool_task_var(chip, param_key='ifp_tie_separation',
                      default_value='0',
                      schelp='maximum distance between tie high/low cells in microns')

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    mainlib = get_mainlib(chip)

    # Set required keys
    for var0, var1 in [('openroad_tielow_cell', 'openroad_tielow_port'),
                       ('openroad_tiehigh_cell', 'openroad_tiehigh_port')]:
        key0 = ['library', mainlib, 'option', 'var', tool, var0]
        key1 = ['library', mainlib, 'option', 'var', tool, var1]
        if chip.valid(*key0):
            chip.add('tool', tool, 'task', task, 'require', ",".join(key1),
                     step=step, index=index)
        if chip.valid(*key1):
            chip.add('tool', tool, 'task', task, 'require', ",".join(key0),
                     step=step, index=index)


def define_ppl_params(chip):
    set_tool_task_var(chip, param_key='ppl_arguments',
                      default_value=[],
                      schelp='additional arguments to pass along to the pin placer.')

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    chip.set('tool', tool, 'task', task, 'file', 'ppl_constraints',
             'pin placement constraints script',
             field='help')

    pdkname = chip.get('option', 'pdk')
    stackup = chip.get('option', 'stackup')

    for key in (['pdk', pdkname, 'var', 'openroad', 'pin_layer_horizontal', stackup],
                ['pdk', pdkname, 'var', 'openroad', 'pin_layer_vertical', stackup]):
        chip.add('tool', tool, 'task', task, 'require', ",".join(key),
                 step=step, index=index)


def define_pdn_params(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    set_tool_task_var(chip, param_key='pdn_enable',
                      default_value='true',
                      schelp='true/false, when true enables power grid generation')

    # Files
    chip.set('tool', tool, 'task', task, 'file', 'pdn_config',
             'list of files to use for power grid generation',
             field='help')


def define_pdn_files(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    targetlibs = get_libraries(chip, 'logic')
    macrolibs = get_libraries(chip, 'macro')

    for libvar, openroadvar in [('openroad_pdngen', 'pdn_config')]:
        if chip.valid('tool', tool, 'task', task, 'file', openroadvar) and \
           chip.get('tool', tool, 'task', task, 'file', openroadvar, step=step, index=index):
            # value already set
            continue

        # copy from libs
        for lib in targetlibs + macrolibs:
            if chip.valid('library', lib, 'option', 'file', libvar):
                for vfile in chip.find_files('library', lib, 'option', 'file', libvar):
                    chip.add('tool', tool, 'task', task, 'file', openroadvar, vfile,
                             step=step, index=index)


def define_pad_params(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    chip.set('tool', tool, 'task', task, 'file', 'padring',
             'script to generate a padring using ICeWall in OpenROAD',
             field='help')

    if chip.get('tool', tool, 'task', task, 'file', 'padring', step=step, index=index):
        chip.add('tool', tool, 'task', task, 'require',
                 ",".join(['tool', tool, 'task', task, 'file', 'padring']),
                 step=step, index=index)


def define_rsz_params(chip):
    set_tool_task_var(chip, param_key='rsz_skip_setup_repair',
                      default_value=False,
                      schelp='skip setup timing repair')
    set_tool_task_var(chip, param_key='rsz_setup_slack_margin',
                      default_value='0.0',
                      schelp='specifies the margin to apply when performing setup repair '
                             'in library timing units')
    set_tool_task_var(chip, param_key='rsz_skip_hold_repair',
                      default_value=False,
                      schelp='skip hold timing repair')
    set_tool_task_var(chip, param_key='rsz_hold_slack_margin',
                      default_value='0.0',
                      schelp='specifies the margin to apply when performing setup repair '
                             'in library timing units')
    set_tool_task_var(chip, param_key='rsz_slew_margin',
                      default_value='0.0',
                      schelp='specifies the amount of margin to apply to max slew repairs '
                             'in percent (0 - 100)')
    set_tool_task_var(chip, param_key='rsz_cap_margin',
                      default_value='0.0',
                      schelp='specifies the amount of margin to apply to max capacitance repairs '
                             'in percent (0 - 100)')
    set_tool_task_var(chip, param_key='rsz_buffer_inputs',
                      default_value='false',
                      schelp='true/false, when true enables adding buffers to the input ports')
    set_tool_task_var(chip, param_key='rsz_buffer_outputs',
                      default_value='false',
                      schelp='true/false, when true enables adding buffers to the output ports')

    set_tool_task_var(chip, param_key='rsz_skip_pin_swap',
                      default_value='true',
                      schelp='true/false, skip pin swap optimization')
    set_tool_task_var(chip, param_key='rsz_skip_gate_cloning',
                      default_value='true',
                      schelp='true/false, skip gate cloning optimization')
    set_tool_task_var(chip, param_key='rsz_repair_tns',
                      default_value='100',
                      schelp='percentage of violating nets to attempt to repair (0 - 100)')

    set_tool_task_var(chip, param_key='rsz_skip_recover_power',
                      default_value=False,
                      schelp='skip power recovery')
    set_tool_task_var(chip, param_key='rsz_recover_power',
                      default_value=100,
                      schelp='percentage of paths to attempt to recover power (0 - 100)')

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    pdkname = chip.get('option', 'pdk')
    stackup = chip.get('option', 'stackup')

    # Set required keys
    for key in (['pdk', pdkname, 'var', 'openroad', 'rclayer_signal', stackup],
                ['pdk', pdkname, 'var', 'openroad', 'rclayer_clock', stackup]):
        chip.add('tool', tool, 'task', task, 'require', ",".join(key),
                 step=step, index=index)


def define_gpl_params(chip):
    set_tool_task_var(chip, param_key='place_density',
                      require=['key'],
                      schelp='global placement density (0.0 - 1.0)')
    set_tool_task_var(chip, param_key='pad_global_place',
                      require=['key'],
                      schelp='global placement cell padding in number of sites')

    set_tool_task_var(chip, param_key='gpl_routability_driven',
                      default_value='true',
                      schelp='true/false, when true global placement will consider the '
                             'routability of the design')
    set_tool_task_var(chip, param_key='gpl_timing_driven',
                      default_value='true',
                      schelp='true/false, when true global placement will consider the '
                             'timing performance of the design')
    set_tool_task_var(chip, param_key='gpl_uniform_placement_adjustment',
                      default_value='0.00',
                      schelp='percent of remaining area density to apply above '
                             'uniform density (0.00 - 0.99)')
    set_tool_task_var(chip, param_key='gpl_enable_skip_io',
                      default_value='true',
                      schelp='true/false, when enabled a global placement is performed without '
                             'considering the impact of the pin placements')


def define_dpo_params(chip):
    set_tool_task_var(chip, param_key='dpo_enable',
                      default_value='true',
                      schelp='true/false, when true the detailed placement optimization '
                             'will be performed')
    set_tool_task_var(chip, param_key='dpo_max_displacement',
                      default_value='0',
                      schelp='maximum cell movement in detailed placement optimization in microns, '
                             '0 will result in the tool default maximum displacement')


def define_dpl_params(chip):
    set_tool_task_var(chip, param_key='pad_detail_place',
                      require=['key'],
                      schelp='detailed placement cell padding in number of sites')

    set_tool_task_var(chip, param_key='dpl_max_displacement',
                      default_value='0',
                      schelp='maximum cell movement in detailed placement in microns, '
                             '0 will result in the tool default maximum displacement')
    set_tool_task_var(chip, param_key='dpl_disallow_one_site',
                      default_value='false',
                      schelp='true/false, disallow single site gaps in detail placement')

    set_tool_task_var(chip, param_key='dpl_use_decap_fillers',
                      default_value='true',
                      schelp='true/false, use decap fillers along with non-decap fillers')


def define_cts_params(chip):
    set_tool_task_var(chip, param_key='cts_distance_between_buffers',
                      default_value='100',
                      schelp='maximum distance between buffers during clock tree synthesis '
                             'in microns')
    set_tool_task_var(chip, param_key='cts_cluster_diameter',
                      default_value='100',
                      schelp='clustering distance to use during clock tree synthesis in microns')
    set_tool_task_var(chip, param_key='cts_cluster_size',
                      default_value='30',
                      schelp='number of instances in a cluster to use during clock tree synthesis')
    set_tool_task_var(chip, param_key='cts_balance_levels',
                      default_value='true',
                      schelp='perform level balancing in clock tree synthesis')
    set_tool_task_var(chip, param_key='cts_obstruction_aware',
                      default_value='true',
                      schelp='make clock tree synthesis aware of obstructions')


def define_grt_params(chip, load_all=False):
    pdkname = chip.get('option', 'pdk')
    stackup = chip.get('option', 'stackup')

    pdk_min_layer = chip.get('pdk', pdkname, 'minlayer', stackup)
    pdk_max_layer = chip.get('pdk', pdkname, 'maxlayer', stackup)

    set_tool_task_var(chip, param_key='grt_setup',
                      default_value='true',
                      schelp='true/false, when true global route is setup')
    set_tool_task_var(chip, param_key='grt_signal_min_layer',
                      default_value=pdk_min_layer,
                      schelp='minimum layer to use for global routing of signals')
    set_tool_task_var(chip, param_key='grt_signal_max_layer',
                      default_value=pdk_max_layer,
                      schelp='maximum layer to use for global routing of signals')
    set_tool_task_var(chip, param_key='grt_clock_min_layer',
                      default_value=pdk_min_layer,
                      schelp='minimum layer to use for global routing of clock nets')
    set_tool_task_var(chip, param_key='grt_clock_max_layer',
                      default_value=pdk_max_layer,
                      schelp='maximum layer to use for global routing of clock nets')
    set_tool_task_var(chip, param_key='grt_macro_extension',
                      default_value='0',
                      schelp='macro extension distance in number of gcells, this can be useful '
                             'when the detailed router needs additional space to avoid DRCs')

    if load_all:
        set_tool_task_var(chip, param_key='grt_use_pin_access',
                          default_value='false',
                          schelp='true/false, when true perform pin access before global routing')
        set_tool_task_var(chip, param_key='grt_overflow_iter',
                          default_value='100',
                          schelp='maximum number of iterations to use in global routing when '
                                 'attempting to solve overflow')
        set_tool_task_var(chip, param_key='grt_allow_congestion',
                          default_value='false',
                          schelp='true/false, when true allow global routing to finish '
                                 'with congestion')
        set_tool_task_var(chip, param_key='grt_allow_overflow',
                          default_value='false',
                          schelp='true/false, when true allow global routing to finish '
                                 'with overflow')


def define_ant_params(chip):
    set_tool_task_var(chip, param_key='ant_iterations',
                      default_value='3',
                      schelp='maximum number of repair iterations to use during antenna repairs')
    set_tool_task_var(chip, param_key='ant_margin',
                      default_value='0',
                      schelp='adds a margin to the antenna ratios (0 - 100)')
    set_tool_task_var(chip, param_key='ant_check',
                      default_value='true',
                      schelp='true/false, flag to indicate whether to check for antenna violations')
    set_tool_task_var(chip, param_key='ant_repair',
                      default_value='true',
                      schelp='true/false, flag to indicate whether to repair antenna violations')


def define_drt_params(chip):
    set_tool_task_var(chip, param_key='drt_disable_via_gen',
                      default_value='false',
                      schelp='true/false, when true turns off via generation in detailed router '
                             'and only uses the specified tech vias')
    set_tool_task_var(chip, param_key='drt_process_node',
                      schelp='when set this specifies to the detailed router the '
                             'specific process node')
    set_tool_task_var(chip, param_key='drt_via_in_pin_bottom_layer',
                      schelp='TODO')
    set_tool_task_var(chip, param_key='drt_via_in_pin_top_layer',
                      schelp='TODO')
    set_tool_task_var(chip, param_key='drt_repair_pdn_vias',
                      schelp='TODO')
    # TODO: This parameter maybe deprecated in favor of drt_repair_pdn_vias
    set_tool_task_var(chip, param_key='drt_via_repair_post_route',
                      default_value='false',
                      schelp='true/false, when true performs a via ripup step after detailed '
                             'routing to remove power vias that are causing DRC violations')

    set_tool_task_var(chip, param_key='detailed_route_default_via',
                      schelp='list of default vias to use for detail routing')
    set_tool_task_var(chip, param_key='detailed_route_unidirectional_layer',
                      schelp='list of layers to treat as unidirectional regardless of '
                             'what the tech lef specifies')

    set_tool_task_var(chip, param_key='drt_report_interval',
                      default_value=5,
                      schelp='reporting interval in steps for generating a DRC report.')


def define_sta_params(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    set_tool_task_var(chip, param_key='sta_early_timing_derate',
                      default_value='0.0',
                      schelp='timing derating factor to use for hold corners')
    set_tool_task_var(chip, param_key='sta_late_timing_derate',
                      default_value='0.0',
                      schelp='timing derating factor to use for setup corners')
    set_tool_task_var(chip, param_key='sta_top_n_paths',
                      default_value='10',
                      schelp='number of paths to report timing for')

    chip.set('tool', tool, 'task', task, 'var', 'power_corner', get_power_corner(chip),
             step=step, index=index, clobber=False)
    chip.add('tool', tool, 'task', task, 'require',
             ','.join(['tool', tool, 'task', task, 'var', 'power_corner']),
             step=step, index=index)
    chip.set('tool', tool, 'task', task, 'var', 'power_corner',
             'corner to use for power analysis',
             field='help')

    add_common_file(chip, 'opensta_generic_sdc', 'sdc/sc_constraints.sdc')


def define_sdc_params(chip):
    set_tool_task_var(chip, param_key='sdc_buffer',
                      schelp='buffer cell to use when auto generating timing constraints')


def define_psm_params(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    set_tool_task_var(chip, param_key='psm_enable',
                      default_value='true',
                      schelp='true/false, when true enables IR drop analysis')

    chip.set('tool', tool, 'task', task, 'var', 'psm_skip_nets',
             'list of nets to skip power grid analysis on',
             field='help')

    chip.set('tool', tool, 'task', task, 'var', 'psm_allow_missing_terminal_nets',
             'list of nets where a missing terminal is acceptable',
             field='help')


def define_fin_params(chip):
    set_tool_task_var(chip, param_key='fin_add_fill',
                      default_value='true',
                      schelp='true/false, when true enables adding fill, '
                             'if enabled by the PDK, to the design')


def define_mpl_params(chip):
    set_tool_task_var(chip, param_key='macro_place_halo',
                      require=['key'],
                      schelp='macro halo to use when performing automated '
                             'macro placement ([x, y] in microns)')
    set_tool_task_var(chip, param_key='macro_place_channel',
                      require=['key'],
                      schelp='macro channel to use when performing automated '
                             'macro placement ([x, y] in microns)')

    set_tool_task_var(chip, param_key='rtlmp_min_instances',
                      schelp='minimum number of instances to use while clustering for '
                             'macro placement')
    set_tool_task_var(chip, param_key='rtlmp_max_instances',
                      schelp='maximum number of instances to use while clustering for '
                             'macro placement')
    set_tool_task_var(chip, param_key='rtlmp_min_macros',
                      schelp='minimum number of macros to use while clustering for macro placement')
    set_tool_task_var(chip, param_key='rtlmp_max_macros',
                      schelp='maximum number of macros to use while clustering for macro placement')
    set_tool_task_var(chip, param_key='rtlmp_max_levels',
                      schelp='maximum depth of physical hierarchical tree')
    set_tool_task_var(chip, param_key='rtlmp_min_aspect_ratio',
                      schelp='Specifies the minimum aspect ratio of its width to height of a '
                             'standard cell cluster')
    set_tool_task_var(chip, param_key='rtlmp_fence',
                      schelp='Defines the global fence bounding box coordinates '
                             '(llx, lly, urx, ury)')
    set_tool_task_var(chip, param_key='rtlmp_bus_planning',
                      schelp='Flag to enable bus planning')
    set_tool_task_var(chip, param_key='rtlmp_target_dead_space',
                      schelp='Specifies the target dead space percentage, which influences '
                             'the utilization of standard cell clusters')

    set_tool_task_var(chip, param_key='rtlmp_area_weight',
                      schelp='Weight for the area of current floorplan')
    set_tool_task_var(chip, param_key='rtlmp_outline_weight',
                      schelp='Weight for violating the fixed outline constraint, meaning that all '
                             'clusters should be placed within the shape of their parent cluster')
    set_tool_task_var(chip, param_key='rtlmp_wirelength_weight',
                      schelp='Weight for half-perimeter wirelength')
    set_tool_task_var(chip, param_key='rtlmp_guidance_weight',
                      schelp='Weight for guidance cost or clusters being placed near specified '
                             'regions if users provide such constraints')
    set_tool_task_var(chip, param_key='rtlmp_fence_weight',
                      schelp='Weight for fence cost, or how far the macro is from zero '
                             'fence violation')
    set_tool_task_var(chip, param_key='rtlmp_boundary_weight',
                      schelp='Weight for the boundary, or how far the hard macro clusters are from '
                             'boundaries. Note that mixed macro clusters are not pushed, thus not '
                             'considered in this cost.')
    set_tool_task_var(chip, param_key='rtlmp_blockage_weight',
                      schelp='Weight for the boundary, or how far the hard macro clusters are '
                             'from boundaries')
    set_tool_task_var(chip, param_key='rtlmp_notch_weight',
                      schelp='Weight for the notch, or the existence of dead space that cannot be '
                             'used for placement & routing')
    set_tool_task_var(chip, param_key='rtlmp_macro_blockage_weight',
                      schelp='Weight for macro blockage, or the overlapping instances of the macro')


def define_ord_params(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    chip.set('tool', tool, 'task', task, 'file', 'global_connect',
             'list of files to use for specifying global connections',
             field='help')

    set_tool_task_var(chip, param_key='ord_abstract_lef_bloat_factor',
                      default_value='10',
                      require=['key'],
                      schelp='Factor to apply when writing the abstract lef')

    set_tool_task_var(chip, param_key='ord_abstract_lef_bloat_layers',
                      default_value='true',
                      require=['key'],
                      schelp='true/false, fill all layers when writing the abstract lef')

    set_tool_task_var(chip, param_key='ord_enable_images',
                      default_value='true',
                      require=['key'],
                      schelp='true/false, enable generating images of the design at the '
                             'end of the task')

    set_tool_task_var(chip, param_key='ord_heatmap_bins_x',
                      default_value='16',
                      require=['key'],
                      schelp='number of X bins to use for heatmap image generation')
    set_tool_task_var(chip, param_key='ord_heatmap_bins_y',
                      default_value='16',
                      require=['key'],
                      schelp='number of Y bins to use for heatmap image generation')

    set_tool_task_lib_var(chip, param_key='scan_chain_cells',
                          default_value=None,
                          schelp='cells to use for scan chain insertion')

    set_tool_task_lib_var(chip, param_key='multibit_ff_cells',
                          default_value=None,
                          schelp='multibit flipflop cells')


def define_ord_files(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)
    targetlibs = get_libraries(chip, 'logic')
    macrolibs = get_libraries(chip, 'macro')

    for libvar, openroadvar in [('openroad_global_connect', 'global_connect')]:
        if chip.valid('tool', tool, 'task', task, 'file', openroadvar) and \
           chip.get('tool', tool, 'task', task, 'file', openroadvar, step=step, index=index):
            # value already set
            continue

        # copy from libs
        for lib in targetlibs + macrolibs:
            if chip.valid('library', lib, 'option', 'file', libvar):
                for vfile in chip.find_files('library', lib, 'option', 'file', libvar):
                    chip.add('tool', tool, 'task', task, 'file', openroadvar, vfile,
                             step=step, index=index)


def define_pex_params(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    chip.set('tool', tool, 'task', task, 'var', 'pex_corners', get_pex_corners(chip),
             step=step, index=index, clobber=False)
    chip.set('tool', tool, 'task', task, 'var', 'pex_corners',
             'list of parasitic extraction corners to use',
             field='help')

    # Auto generated file
    chip.set('tool', tool, 'task', task, 'file', 'parasitics',
             'file used to specify the parasitics for estimation',
             field='help')


def set_reports(chip, reports):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    # supported reports
    supported = (
        "setup",
        "hold",
        "unconstrained",
        "clock_skew",
        "drv_violations",
        "fmax",
        "power",
        "check_setup",
        "placement_density",
        "routing_congestion",
        "power_density",
        "ir_drop",
        "clock_placement",
        "clock_trees",
        "optimization_placement",
        "module_view"
    )

    chip.set('tool', tool, 'task', task, 'var', 'reports',
             'list of reports and images to generate',
             field='help')

    chip.set('tool', tool, 'task', task, 'var', 'reports', [],
             step=step, index=index)

    def check_enabled(type):
        for key in (('tool', tool, 'task', task, 'var', f'skip_{type}'),
                    ('option', 'var', f'openroad_skip_{type}')):
            if chip.valid(*key):
                if chip.get(*key, field='pernode').is_never():
                    if chip.get(*key) == ["true"]:
                        return False
                    elif chip.get(*key, step=step, index=index) == ["true"]:
                        return False
        return True

    for report in reports:
        if report not in supported:
            raise ValueError(f'{report} is not supported')
        if check_enabled(report):
            chip.add('tool', tool, 'task', task, 'var', 'reports', report,
                     step=step, index=index)


def set_pnr_inputs(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    design = chip.top()

    if f'{design}.sdc' in input_provides(chip, step, index):
        chip.add('tool', tool, 'task', task, 'input', design + '.sdc',
                 step=step, index=index)
    elif chip.valid('input', 'constraint', 'sdc') and \
            chip.get('input', 'constraint', 'sdc', step=step, index=index):
        chip.add('tool', tool, 'task', task, 'require', 'input,constraint,sdc',
                 step=step, index=index)

    if f'{design}.odb' in input_provides(chip, step, index):
        chip.add('tool', tool, 'task', task, 'input', design + '.odb',
                 step=step, index=index)
    elif f'{design}.def' in input_provides(chip, step, index):
        chip.add('tool', tool, 'task', task, 'input', design + '.def',
                 step=step, index=index)
    elif chip.valid('input', 'layout', 'def') and \
            chip.get('input', 'layout', 'def', step=step, index=index):
        chip.add('tool', tool, 'task', task, 'require', 'input,layout,def',
                 step=step, index=index)


def set_pnr_outputs(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    design = chip.top()

    chip.add('tool', tool, 'task', task, 'output', design + '.sdc', step=step, index=index)
    chip.add('tool', tool, 'task', task, 'output', design + '.vg', step=step, index=index)
    chip.add('tool', tool, 'task', task, 'output', design + '.def', step=step, index=index)
    chip.add('tool', tool, 'task', task, 'output', design + '.odb', step=step, index=index)


def get_library_timing_keypaths(chip, lib):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')

    delaymodel = chip.get('asic', 'delaymodel', step=step, index=index)
    keypaths = {}
    for constraint in chip.getkeys('constraint', 'timing'):
        corners = chip.get('constraint', 'timing', constraint, 'libcorner', step=step, index=index)
        for corner in corners:
            if chip.valid('library', lib, 'output', corner, delaymodel):
                keypaths[constraint] = ('library', lib, 'output', corner, delaymodel)

        if constraint not in keypaths:
            keypaths[constraint] = ('library', lib, 'output', corners[0], delaymodel)
    return keypaths


def get_pex_corners(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')

    corners = set()
    for constraint in chip.getkeys('constraint', 'timing'):
        pexcorner = chip.get('constraint', 'timing', constraint, 'pexcorner',
                             step=step, index=index)
        if pexcorner:
            corners.add(pexcorner)

    return list(corners)


def _get_constraint_by_check(chip, check):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')

    for constraint in chip.getkeys('constraint', 'timing'):
        if check in chip.get('constraint', 'timing', constraint, 'check',
                             step=step, index=index):
            return constraint

    # if not specified, just pick the first constraint available
    return chip.getkeys('constraint', 'timing')[0]


def get_power_corner(chip):
    return _get_constraint_by_check(chip, "power")


def get_setup_corner(chip):
    return _get_constraint_by_check(chip, "setup")


def build_pex_corners(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    pdkname = chip.get('option', 'pdk')
    stackup = chip.get('option', 'stackup')

    corners = {}
    for constraint in chip.getkeys('constraint', 'timing'):
        pexcorner = chip.get('constraint', 'timing', constraint, 'pexcorner',
                             step=step, index=index)

        if not pexcorner:
            continue
        corners[constraint] = pexcorner

    default_corner = get_setup_corner(chip)
    if default_corner in corners:
        corners[None] = corners[default_corner]

    chip.set('tool', tool, 'task', task, 'file', 'parasitics',
             os.path.join(chip.getworkdir(step=step, index=index),
                          'inputs',
                          'sc_parasitics.tcl'),
             step=step, index=index, clobber=True)
    chip.set('tool', tool, 'task', task, 'file', 'parasitics', False, field='copy')

    with open(chip.get('tool', tool, 'task', task, 'file', 'parasitics',
                       step=step, index=index)[0], 'w') as f:
        for constraint, pexcorner in corners.items():
            if chip.valid('pdk', pdkname, 'pexmodel', tool, stackup, pexcorner):
                pex_source_file = chip.find_files(
                    'pdk',
                    pdkname,
                    'pexmodel',
                    tool,
                    stackup,
                    pexcorner)[0]
                if not pex_source_file:
                    continue

                corner_pex_template = utils.get_file_template(pex_source_file)
                pex_template = utils.get_file_template(
                    'pex.tcl',
                    root=os.path.join(os.path.dirname(__file__),
                                      'templates'))

                if not pex_template:
                    continue

                if constraint is None:
                    constraint = "default"
                    corner_specification = ""
                else:
                    corner_specification = f"-corner {constraint}"

                f.write(pex_template.render(
                    constraint=constraint,
                    pexcorner=pexcorner,
                    source=pex_source_file,
                    pex=corner_pex_template.render({"corner": corner_specification})
                ))
                f.write('\n')



import os

from siliconcompiler import utils
from siliconcompiler.tools.netgen import count_lvs
from siliconcompiler import sc_open
from siliconcompiler.tools._common import get_tool_task, record_metric


def setup(chip):
    '''
    Perform LVS on the supplied netlists
    '''

    tool = 'netgen'
    refdir = 'tools/' + tool
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)

    # magic used for drc and lvs
    script = 'sc_lvs.tcl'

    chip.set('tool', tool, 'exe', tool)
    chip.set('tool', tool, 'vswitch', '-batch')
    chip.set('tool', tool, 'version', '>=1.5.192', clobber=False)
    chip.set('tool', tool, 'format', 'tcl')

    chip.set('tool', tool, 'task', task, 'threads', utils.get_cores(chip),
             step=step, index=index, clobber=False)
    chip.set('tool', tool, 'task', task, 'refdir', refdir,
             step=step, index=index,
             package='siliconcompiler', clobber=False)
    chip.set('tool', tool, 'task', task, 'script', script,
             step=step, index=index, clobber=False)

    # set options
    options = []
    options.append('-batch')
    options.append('source')
    chip.set('tool', tool, 'task', task, 'option', options, step=step, index=index, clobber=False)

    design = chip.top()
    chip.add('tool', tool, 'task', task, 'input', f'{design}.spice', step=step, index=index)
    if chip.valid('input', 'netlist', 'verilog'):
        chip.add('tool', tool, 'task', task, 'require',
                 ','.join(['input', 'netlist', 'verilog']),
                 step=step, index=index)
    else:
        chip.add('tool', tool, 'task', task, 'input', f'{design}.vg', step=step, index=index)

    # Netgen doesn't have a standard error prefix that we can grep for, but it
    # does print all errors to stderr, so we can redirect them to <step>.errors
    # and use that file to count errors.
    chip.set('tool', tool, 'task', task, 'stderr', 'suffix', 'errors', step=step, index=index)

    chip.set('tool', tool, 'task', task, 'regex', 'warnings', '^Warning:',
             step=step, index=index, clobber=False)


################################
# Post_process (post executable)
################################
def post_process(chip):
    ''' Tool specific function to run after step execution

    Reads error count from output and fills in appropriate entry in metrics
    '''
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    design = chip.top()

    with sc_open(f'{step}.errors') as f:
        errors = len([line for line in f.readlines() if not line.startswith("Note:")])
    record_metric(chip, step, index, 'errors', errors, f'{step}.errors')

    # Export metrics
    lvs_report = f'reports/{design}.lvs.json'
    if not os.path.isfile(lvs_report):
        chip.logger.warning('No LVS report generated. Netgen may have encountered errors.')
        return

    lvs_failures = count_lvs.count_LVS_failures(lvs_report)

    # We don't count top-level pin mismatches as errors b/c we seem to get
    # false positives for disconnected pins. Report them as warnings
    # instead, the designer can then take a look at the full report for
    # details.
    pin_failures = lvs_failures[3]
    errors = lvs_failures[0] - pin_failures
    record_metric(chip, step, index, 'drcs', errors, lvs_report)
    record_metric(chip, step, index, 'warnings', pin_failures, lvs_report)



# Copyright 2024 Silicon Compiler Authors. All Rights Reserved.

# Standard Modules
import base64
import io
import json
import gzip
import os
import stat
import sys
import tarfile
import tempfile
import textwrap

from datetime import datetime

import siliconcompiler
from siliconcompiler.apps._common import UNSET_DESIGN
from siliconcompiler import SiliconCompilerError
from siliconcompiler import utils


def make_bytes(data):
    if isinstance(data, bytes):
        return data
    return data.encode('utf-8')


def compress(data):
    return gzip.compress(make_bytes(data))


def convert_base64(data):
    return base64.b64encode(make_bytes(data))


def wrap_text(data):
    if isinstance(data, bytes):
        data = data.decode('utf-8')
    return textwrap.wrap(data)


###########################
def main():
    progname = "summarize"
    description = """
    ------------------------------------------------------------
    Utility script to print job record information from a manifest
    needed to replay that manifest.
    ------------------------------------------------------------
    """
    # Create a base chip class.
    chip = siliconcompiler.Chip(UNSET_DESIGN)

    # Read command-line inputs and generate Chip objects to run the flow on.
    try:
        file_arg = {
            'metavar': '<file>',
            'help': 'Path to generate replay file to.',
            'default': 'replay.sh',
            'sc_print': True
        }
        args = chip.create_cmdline(
            progname,
            description=description,
            switchlist=['-cfg',
                        '-jobname',
                        '-loglevel'],
            additional_args={
                '-file': file_arg
            })
    except SiliconCompilerError:
        return 1
    except Exception as e:
        chip.logger.error(e)
        return 1

    design = chip.get('design')
    if design == UNSET_DESIGN:
        chip.logger.error('Design not loaded')
        return 1

    # Print Job Summary
    jobname = chip.get('option', 'jobname')
    pythonpackages = chip.get('record', 'pythonpackage', job=jobname)

    pythonversion = set()
    nodes = set()
    for version, step, index in chip.schema._getvals('history', jobname, 'record', 'pythonversion'):
        pythonversion.add(version)
        nodes.add((step, index))

    if len(pythonversion) > 1:
        chip.logger.warning(f"More than one python version detected: {', '.join(pythonversion)}")
    pythonversion = list(pythonversion)[0]

    tools = {}
    tool_versions = []
    for step, index in nodes:
        toolpath = chip.get('record', 'toolpath', job=jobname, step=step, index=index)
        toolversion = chip.get('record', 'toolversion', job=jobname, step=step, index=index)

        if toolpath is None:
            continue

        tools.setdefault(toolpath, set()).add(toolversion)
        if toolversion:
            tool = chip.get('flowgraph', chip.get('option', 'flow'), step, index, 'tool')
            tool_versions.append(
                ((step, index), tool, toolversion)
            )

    print("SUMMARY       :")
    print(f"design        : {chip.design}")
    print(f"pythonversion : {pythonversion}")

    print("Python packages requires:")
    for pkg in sorted(pythonpackages):
        print(f"  {pkg}")

    print("Tool requirements:")
    tool_len = max([len(os.path.basename(tool)) for tool, _ in tools.items()])
    for tool, version in tools.items():
        print(f"  {os.path.basename(tool):<{tool_len}}: {', '.join(version)}")

    path = os.path.abspath(args['file'])
    os.makedirs(os.path.dirname(path), exist_ok=True)

    starttimes = set()
    for starttime, step, index in chip.schema._getvals('history', jobname, 'record', 'starttime'):
        starttimes.add(datetime.strptime(starttime, '%Y-%m-%d %H:%M:%S'))
    starttime = min(starttimes).strftime('%Y-%m-%d %H:%M:%S')

    with io.StringIO() as fd:
        fd.write(utils.get_file_template('replay/requirements.txt').render(
            design=chip.design,
            jobname=jobname,
            date=starttime,
            pkgs=pythonpackages
        ))
        fd.flush()
        requirements_file = fd.getvalue()

    with tempfile.TemporaryDirectory() as collect:
        chip.collect(directory=collect, verbose=True, exclude_packages=['siliconcompiler'])

        with io.BytesIO() as fd:
            with tarfile.open(fileobj=fd, mode='w:gz') as tar:
                tar.add(collect, arcname='')

            fd.flush()
            collect_files = convert_base64(fd.getvalue())

    with io.StringIO() as fd:
        fd.write(utils.get_file_template('replay/replay.py.j2').render(
            design=chip.design,
            jobname=jobname,
            date=starttime,
            src_file=wrap_text(collect_files),
            tool_versions=sorted(tool_versions)
        ))
        fd.flush()
        script = convert_base64(compress(fd.getvalue()))

    manifest = convert_base64(compress(json.dumps(chip.schema.cfg, indent=2)))

    tool_info = []
    for tool, version in tools.items():
        tool_info.append(f"{os.path.basename(tool):<{tool_len}}: {', '.join(version)}")

    description = f"Replay for {chip.design} / {jobname}\nRun on: {starttime}"

    with open(path, 'w', encoding='utf-8') as wf:
        wf.write(utils.get_file_template('replay/setup.sh').render(
            design=chip.design,
            jobname=jobname,
            date=starttime,
            description=description,
            pythonversion=pythonversion,
            requirements=requirements_file.splitlines(),
            script=wrap_text(script),
            manifest=wrap_text(manifest),
            tools=tool_info
        ))

    permissions = stat.S_IMODE(os.lstat(path).st_mode)
    os.chmod(path, permissions | stat.S_IXUSR)

    return 0


#########################
if __name__ == "__main__":
    sys.exit(main())



import re

from siliconcompiler.tools.yosys import setup as setup_tool
from siliconcompiler.tools.yosys.syn_asic import setup_asic, prepare_synthesis_libraries
from siliconcompiler import sc_open
from siliconcompiler.tools._common import get_tool_task, record_metric, input_provides


def setup(chip):
    '''
    Perform logical equivalence checks
    '''

    # Generic tool setup.
    setup_tool(chip)

    # Generic setup.
    setup_asic(chip)

    tool = 'yosys'
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)
    design = chip.top()

    # Set yosys script path.
    chip.set('tool', tool, 'task', task, 'script', 'sc_lec.tcl',
             step=step, index=index, clobber=False)

    # Input/output requirements.
    if f"{design}.lec.vg" in input_provides(chip, step, index):
        chip.set('tool', tool, 'task', task, 'input', design + '.lec.vg',
                 step=step, index=index)
    elif f"{design}.vg" in input_provides(chip, step, index):
        chip.set('tool', tool, 'task', task, 'input', design + '.vg',
                 step=step, index=index)
    else:
        chip.add('tool', tool, 'task', task, 'require', 'input,netlist,verilog',
                 step=step, index=index)

    chip.set('tool', tool, 'task', task, 'var', 'induction_steps', '10',
             step=step, index=index, clobber=False)
    chip.set('tool', tool, 'task', task, 'var', 'induction_steps',
             'Number of induction steps for yosys equivalence checking',
             field='help')


def pre_process(chip):
    prepare_synthesis_libraries(chip)


##################################################
def post_process(chip):
    ''' Tool specific function to run after step execution
    '''

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')

    with sc_open(step + ".log") as f:
        for line in f:
            if line.endswith('Equivalence successfully proven!\n'):
                record_metric(chip, step, index, 'drvs', 0, step + ".log")
                continue

            errors = re.search(r'Found a total of (\d+) unproven \$equiv cells.', line)
            if errors is not None:
                num_errors = int(errors.group(1))
                record_metric(chip, step, index, 'drvs', num_errors, step + ".log")



# Copyright 2023 Silicon Compiler Authors. All Rights Reserved.
import sys
import os
import siliconcompiler
import tarfile
import json
from siliconcompiler.scheduler import _runtask, _executenode
from siliconcompiler.issue import generate_testcase
from siliconcompiler.tools._common import get_tool_task


def main():
    progname = "sc-issue"
    chip = siliconcompiler.Chip(progname)
    switchlist = ['-cfg',
                  '-arg_step',
                  '-arg_index',
                  '-tool_task_var',
                  '-tool_task_option',
                  '-loglevel',
                  '-use']
    description = """
-----------------------------------------------------------
Restricted SC app that generates a sharable testcase from a
failed flow or runs an issue generated with this program.

To generate a testcase, use:
    sc-issue -cfg <stepdir>/outputs/<design>.pkg.json

    or include a different step/index than what the cfg_file is pointing to:
    sc-issue -cfg <otherdir>/outputs/<design>.pkg.json -arg_step <step> -arg_index <index>

    or include specific libraries while excluding others:
    sc-issue -cfg <stepdir>/outputs/<design>.pkg.json -exclude_libraries -add_library sram -add_library gpio

To run a testcase, use:
    sc-issue -run -file sc_issue_<...>.tar.gz
-----------------------------------------------------------
"""  # noqa E501

    issue_arguments = {
        '-exclude_libraries': {'action': 'store_true',
                               'help': 'flag to ensure libraries are excluded in the testcase',
                               'sc_print': False},
        '-exclude_pdks': {'action': 'store_true',
                          'help': 'flag to ensure pdks are excluded in the testcase',
                          'sc_print': False},
        '-hash_files': {'action': 'store_true',
                        'help': 'flag to hash the files in the schema before generating '
                                'the manifest',
                        'sc_print': False},

        '-run': {'action': 'store_true',
                 'help': 'run a provided testcase',
                 'sc_print': False},

        '-add_library': {'action': 'append',
                         'help': 'library to include in the testcase, if not provided all '
                                 'libraries will be added according to the -exclude_libraries flag',
                         'metavar': '<library>',
                         'sc_print': False},
        '-add_pdk': {'action': 'append',
                     'help': 'pdk to include in the testcase, if not provided all libraries '
                             'will be added according to the -exclude_pdks flag',
                     'metavar': '<pdk>',
                     'sc_print': False},

        '-file': {'help': 'filename for the generated testcase',
                  'metavar': '<file>',
                  'sc_print': False},
    }

    try:
        switches = chip.create_cmdline(progname,
                                       switchlist=switchlist,
                                       description=description,
                                       additional_args=issue_arguments)
    except Exception as e:
        chip.logger.error(e)
        return 1

    if not switches['run']:
        step = chip.get('arg', 'step')
        index = chip.get('arg', 'index')

        if not step:
            chip.logger.error('Unable to determine step from manifest')
        if not index:
            chip.logger.error('Unable to determine index from manifest')
        if not step or not index:
            # Exit out
            return 1

        generate_testcase(chip,
                          step,
                          index,
                          switches['file'],
                          include_pdks=not switches['exclude_pdks'],
                          include_specific_pdks=switches['add_pdk'],
                          include_libraries=not switches['exclude_libraries'],
                          include_specific_libraries=switches['add_library'],
                          hash_files=switches['hash_files'])

        return 0

    if switches['run']:
        if not switches['file']:
            raise ValueError('-file must be provided')

        test_dir = os.path.basename(switches['file']).split('.')[0]
        with tarfile.open(switches['file'], 'r:gz') as f:
            f.extractall(path='.')

        chip.read_manifest(f'{test_dir}/orig_manifest.json')

        with open(f'{test_dir}/issue.json', 'r') as f:
            issue_info = json.load(f)

        # Report major information
        sc_version = issue_info['version']['sc']
        if 'commit' in issue_info['version']['git']:
            sc_source = issue_info['version']['git']['commit']
        else:
            sc_source = 'installed'

        chip.logger.info(f"Design: {chip.design}")
        chip.logger.info(f"SiliconCompiler version: {sc_version} / {sc_source}")
        chip.logger.info(f"Schema version: {issue_info['version']['schema']}")
        chip.logger.info(f"Python version: {issue_info['python']['version']}")

        sc_machine = issue_info['machine']
        sc_os = sc_machine['system']
        if 'distro' in sc_machine:
            sc_os += f" {sc_machine['distro']}"
        sc_os += f" {sc_machine['osversion']} {sc_machine['arch']} {sc_machine['kernelversion']}"
        chip.logger.info(f"Machine: {sc_os}")
        chip.logger.info(f"Date: {issue_info['date']}")

        step = issue_info['run']['step']
        index = issue_info['run']['index']

        chip.set('arg', 'step', step)
        chip.set('arg', 'index', index)
        tool, task = get_tool_task(chip, step, index)
        chip.logger.info(f'Preparing run for {step}{index} - {tool}/{task}')

        # Modify run environment to point to extracted files
        builddir_key = ['option', 'builddir']
        new_builddir = f"{test_dir}/{chip.get(*builddir_key)}"
        chip.logger.info(f"Changing {builddir_key} to '{new_builddir}'")
        chip.set(*builddir_key, new_builddir)

        # Run task
        # Rerun setup task, assumed to be running in its own thread so
        # multiprocess is not needed
        flow = chip.get('option', 'flow')
        _runtask(chip, flow, step, index, _executenode, replay=True)

        return 0


#########################
if __name__ == "__main__":
    sys.exit(main())
